#ifndef __XMP_BuildInfo_h__
#define __XMP_BuildInfo_h__ 1

/* --------------------------------------------------------------------------------------------- */
/* ** IMPORTANT ** This file must be usable by strict ANSI C compilers. No "//" comments, etc.   */
/* --------------------------------------------------------------------------------------------- */

/*
// =================================================================================================
// Copyright 2002-2007 Adobe Systems Incorporated
// All Rights Reserved.
//
// NOTICE:  Adobe permits you to use, modify, and distribute this file in accordance with the terms
// of the Adobe license agreement accompanying it.
// =================================================================================================
*/


/*
// =================================================================================================
// This file provides build numbers that are set automatically by the external build system. This is
// in contrast to the manual version numbers of XMP_ChangeLog.h.
//
// The value of kXMP_BuildDate is set to the date of the build, in some reasonable format. It would
// be nice if the month used a name instead of a number and if the year has 4 digits. But it isn't
// worth going to great lengths for this, a numeric date is OK.

// The value of kXMP_BuildNumber is set to some number that is meaningful to the build system, such
// as a Perforce changelist number.
// =================================================================================================
*/

#define kXMP_Copyright Copyright (c) 2002-2007, Adobe Systems Incorporated
#define kXMP_CopyrightStr "Copyright (c) 2002-2007, Adobe Systems Incorporated"

#endif	/* __XMP_BuildInfo_h__ */
